##
#  File:           IMGTTargetProvider.py
#  Date:           5-Jul-2021 jdw
#
#  Updated:
#   14-Mar-2023 dwp  Add timeout to IMGT data file fetch
#    3-Jul-2023 aae  imgt.org no longer supports http
#
##
"""
Accessors for IMGT target annotations.
"""

import datetime
import glob
import gzip
import logging
import os.path
import time

from rcsb.utils.io.FileUtil import FileUtil
from rcsb.utils.io.MarshalUtil import MarshalUtil
from rcsb.utils.io.StashableBase import StashableBase
from rcsb.utils.taxonomy.TaxonomyProvider import TaxonomyProvider

logger = logging.getLogger(__name__)


class IMGTTargetProvider(StashableBase):
    """Accessors for IMGT target annotations."""

    def __init__(self, cachePath, useCache, **kwargs):
        #
        self.__cachePath = cachePath
        self.__dirName = "IMGT-targets"
        imgtDumpUrl = kwargs.get("IMGTDumpUrl", None)
        super(IMGTTargetProvider, self).__init__(self.__cachePath, [self.__dirName])
        self.__dirPath = os.path.join(self.__cachePath, self.__dirName)
        self.__version = None
        #
        self.__mU = MarshalUtil(workPath=self.__dirPath)
        self.__imgtD = self.__reload(self.__dirPath, useCache=useCache, imgtDumpUrl=imgtDumpUrl)
        #

    def testCache(self, minCount=1000):
        if self.__imgtD and "chains" in self.__imgtD and len(self.__imgtD["chains"]) > minCount:
            return True
        else:
            return False

    def getVersion(self):
        return self.__version

    def getChains(self):
        return self.__imgtD["chains"]

    def __reload(self, dirPath, useCache=False, imgtDumpUrl=None, testList=None, maxCount=None):
        imgtD = {}
        startTime = time.time()

        fU = FileUtil(timeout=300)
        fU.mkdir(dirPath)
        #
        imgtDataPath = os.path.join(self.__dirPath, "imgt-data.json")
        #
        logger.info("useCache %r imgtFeaturePath %r", useCache, imgtDataPath)
        if useCache and self.__mU.exists(imgtDataPath):
            imgtD = self.__mU.doImport(imgtDataPath, fmt="json")
            self.__version = imgtD["version"]
        else:
            imgtDumpUrl = imgtDumpUrl if imgtDumpUrl else "https://www.imgt.org/download/3Dstructure-DB/IMGT3DFlatFiles.tgz"
            imgtReadmeUrl = "https://www.imgt.org/download/3Dstructure-DB/RELEASE"
            imgtDumpFileName = fU.getFileName(imgtDumpUrl)
            imgtDumpPath = os.path.join(dirPath, imgtDumpFileName)
            imgtReleasePath = os.path.join(dirPath, "IMGT-release.txt")
            _, fn = os.path.split(imgtDumpUrl)
            imgtFlatFilePath = os.path.join(self.__dirPath, fn[:-4])
            #
            logger.info("Fetching url %s path %s", imgtDumpUrl, imgtDumpPath)
            ok1 = fU.get(imgtDumpUrl, imgtDumpPath)
            ok2 = fU.get(imgtReadmeUrl, imgtReleasePath)
            fU.unbundleTarfile(imgtDumpPath, dirPath=dirPath)
            logger.info("Completed fetch (%r) at %s (%.4f seconds)", ok1 and ok2, time.strftime("%Y %m %d %H:%M:%S", time.localtime()), time.time() - startTime)
            # ---
            readmeLines = self.__mU.doImport(imgtReleasePath, fmt="list")
            self.__version = readmeLines[0].strip() if readmeLines else None
            logger.info("IMGT version %r", self.__version)
            # ---
            chainD, rawD = self.__imgtFlatFileProcessor(imgtFlatFilePath, maxCount=maxCount, testList=testList)
            # ---
            tS = datetime.datetime.now().isoformat()
            # vS = datetime.datetime.now().strftime("%Y-%m-%d")
            if testList:
                imgtD = {"version": self.__version, "date": tS, "chains": chainD, "raw": rawD}
            else:
                imgtD = {"version": self.__version, "date": tS, "chains": chainD}
            ok = self.__mU.doExport(imgtDataPath, imgtD, fmt="json", indent=3)
            logger.info("Completed flatfile prep (%r) at %s (%.4f seconds)", ok, time.strftime("%Y %m %d %H:%M:%S", time.localtime()), time.time() - startTime)
        return imgtD

    def exportFasta(self, withGaps=False):
        """
        Example:
            The IMGT/GENE-DB FASTA header contains 15 fields separated by '|':

            1. IMGT/LIGM-DB accession number(s)
            2. IMGT gene and allele name
            3. species (may be followed by an "_" and the name of the strain, breed or isolate, if defined)
            4. IMGT gene and allele functionality
            5. exon(s), region name(s), or extracted label(s)
            6. start and end positions in the IMGT/LIGM-DB accession number(s)
            7. number of nucleotides in the IMGT/LIGM-DB accession number(s)
            8. codon start, or 'NR' (not relevant) for non coding labels
            9. +n: number of nucleotides (nt) added in 5' compared to the corresponding label extracted from IMGT/LIGM-DB
            10. +n or -n: number of nucleotides (nt) added or removed in 3' compared to the corresponding label extracted from IMGT/LIGM-DB
            11. +n, -n, and/or nS: number of added, deleted, and/or substituted nucleotides to correct sequencing errors, or 'not corrected' if non corrected sequencing errors
            12. number of amino acids (AA): this field indicates that the sequence is in amino acids
            13. number of characters in the sequence: nt (or AA)+IMGT gaps=total
            14. partial (if it is)
            15. reverse complementary (if it is)

        """
        # --
        fU = FileUtil()
        fU.mkdir(self.__dirPath)
        if withGaps:
            imgtTargetUrl = "https://www.imgt.org/download/GENE-DB/IMGTGENEDB-ReferenceSequences.fasta-AA-WithGaps-F+ORF+inframeP"
        else:
            imgtTargetUrl = "https://www.imgt.org/download/GENE-DB/IMGTGENEDB-ReferenceSequences.fasta-AA-WithoutGaps-F+ORF+inframeP"
        imgtTargetFileName = fU.getFileName(imgtTargetUrl)
        rawFastaPath = os.path.join(self.__dirPath, imgtTargetFileName)
        # --
        logger.debug("Fetching url %s path %s", imgtTargetUrl, rawFastaPath)
        ok = fU.get(imgtTargetUrl, rawFastaPath)
        logger.info("Fetch status (%r) url %s path %s", ok, imgtTargetUrl, rawFastaPath)
        # --
        fastaPath = os.path.join(self.__dirPath, "imgt-reference.fa")
        taxonPath = os.path.join(self.__dirPath, "imgt-reference-taxon.tdd")
        tP = TaxonomyProvider(cachePath=self.__cachePath, useCache=True)
        ok = tP.testCache()
        if not ok:
            tP = TaxonomyProvider(cachePath=self.__cachePath, useCache=False)

        rawQD = self.__mU.doImport(rawFastaPath, fmt="fasta", commentStyle="default")
        oD = {}
        taxonL = []
        for queryId, sD in rawQD.items():
            qL = queryId.split("|")
            tL = qL[2].split("_")
            taxName = tL[0]
            taxVar = tL[1].replace(" ", "_") if len(tL) > 1 else None
            taxId = tP.getTaxId(taxName)
            if taxId:
                tD = {"seqId": qL[0], "imgtGene": qL[1], "functionality": qL[3], "labels": qL[4], "taxId": taxId}
                if taxVar:
                    tD["taxVar"] = taxVar
                sD.update(tD)
            else:
                logger.info("Unknown taxonomy %r (taxName=%r)", queryId, taxName)
            sD["sequence"].replace(".", "-")
            seqId = ""
            cL = []
            for k, v in sD.items():
                if k in ["sequence"]:
                    continue
                cL.append(str(v))
                cL.append(str(k))
            seqId = "|".join(cL)
            oD[seqId] = sD
            taxonL.append("%s\t%s" % (seqId, taxId))
        #
        ok1 = self.__mU.doExport(taxonPath, taxonL, fmt="list")
        ok2 = self.__mU.doExport(fastaPath, oD, fmt="fasta", makeComment=True)
        return ok1 and ok2

    def __imgtFlatFileProcessor(self, flatFilePath, maxCount=None, testList=None):
        chainD = {}
        rawD = {}
        failures = []
        idList = []
        ic = 0
        filePattern = os.path.join(flatFilePath, "*.pdb.gz")
        logger.info("Collecting flat files with pattern %r", filePattern)
        for fp in glob.glob(filePattern):
            ic += 1
            if maxCount and ic > maxCount:
                break
            logger.debug("Processing file %r", fp)
            _, fn = os.path.split(fp)
            pdbId = fn[5:9].lower()
            if testList and pdbId not in testList:
                continue
            idList.append(pdbId)
            cD = {}
            tmpD = {}
            with gzip.open(fp, "rb") as ifh:
                try:
                    cD, tmpD = self.__imgtRemarkParser(pdbId, ifh)
                except Exception as e:
                    failures.append(pdbId)
                    logger.exception("Failing for %r with %s", pdbId, str(e))
                    continue
            # --
            chainD.update(cD)
            rawD[pdbId] = tmpD
        #
        logger.info("ID List (%d)", len(set(idList)))
        sL = list(rawD.keys())
        logger.info("Successes (%d) chains (%d)", len(sL), len(chainD))
        logger.info("Exceptions (%d) %r", len(failures), failures)
        mL = list(set(idList) - set(sL))
        logger.info("Missing (%d) %r", len(mL), mL)
        #
        return chainD, rawD

    def __imgtRemarkParser(self, pdbId, ifh):
        """IMGT REMARK 410 Parser

        Args:
            pdbId (str): input PDB ID [description]
            ifh (obj): input file handle

        Returns:
            dict: content dictionary of parsed details
        """
        sD = {
            "IMGT protein name": {"section": "proteins"},
            "ligand(s)": {"section": "ligands"},
            "Chain ID  ": {"section": "chains"},
        }
        pD = {
            "Chain ID  ": {"ky": "chain_data", "action": "appendAll"},
            #
            "ligand(s)": {"ky": "ligands", "action": "appendLine"},
            "IMGT protein name": {"ky": "proteinName", "action": "appendLine"},
            "IMGT receptor type": {"ky": "receptorType", "action": "appendLine"},
            "IMGT receptor description": {"ky": "receptorDescription", "action": "appendLine"},
            "Species": {"ky": "species", "action": "appendLine"},
            "Chain ID": {"ky": "chain_ids", "action": "appendLine"},
            #
        }
        cD = {}
        oD = {}
        curSection = None
        action = None
        curKy = None
        curSection = None
        curChain = None
        for ul in ifh.readlines():
            line = ul.decode("utf-8")
            if not line.startswith("REMARK 410 "):
                continue
            #
            curLine = line[11:-1]

            for section, sectionD in sD.items():
                if curLine.startswith(section):
                    logger.debug("%r Detected section %r", pdbId, section)
                    curSection = sectionD["section"]
                    first = True
                    break
            #
            for label, labelD in pD.items():
                if curLine.startswith(label):
                    logger.debug("%r detected label %r", pdbId, label)
                    curKy = labelD["ky"]
                    action = labelD["action"]
                    first = True
                    break
            #
            if action == "appendLine":
                if first:
                    first = False
                    logger.debug("Skipped %r", curKy)
                    continue
                logger.debug(">> SECTION %r KEY %r Adding %r", curSection, curKy, curLine.strip())
                oD.setdefault(curSection, {}).setdefault(curKy, []).append(curLine.strip())
            elif action == "appendAll":
                if first:
                    tL = [t for t in curLine.split(" ") if t]
                    curChain = tL[2]
                    first = False
                    logger.debug("%r current chain key %r", pdbId, curChain)
                    continue
                oD.setdefault(curSection, {}).setdefault(curChain, []).append(curLine)
        # --  rD raw extracted REM 410 content
        #  Post-process the domain annotations and alignments

        for chId, cL in oD["chains"].items() if "chains" in oD else {}:
            logger.debug("%r chainId %r (%d)", pdbId, chId, len(cL))
            tD = {}
            tD["description"] = self.__getField(cL, label="IMGT chain description  ")
            tD["domains"] = self.__splitDomains(pdbId, cL)
            #
            aD = self.__getAlignment(pdbId, cL)
            if aD and (len(aD["alignMapDL"]) == len(tD["domains"])):
                aL = aD["alignMapDL"]
                for ii, dD in enumerate(tD["domains"].values()):
                    if aL and len(aL) > ii:
                        dD["alignment"] = aL[ii]
            # --
            #  Integrate raw "proteins"  content
            if "proteins" in oD:
                paD = self.__getProteinAnnotations(chId, oD["proteins"])
                logger.debug("paD %r", paD)
                tD.update(paD)
            # --
            cD[chId] = tD
        #
        return cD, oD
        #

    def __getProteinAnnotations(self, chainId, pLD):
        """
        Example:

         "proteins": {
            "proteinName": [
               "IgG4 Sigma1 Fc"
            ],
            "receptorType": [
               "IG"
            ],
            "receptorDescription": [
               "FUSION-[TNFRSF1B]2-FC-GAMMA-1"
            ],
            "species": [
               "Homo sapiens (human)"
            ],
            "chain_ids": [
               "5w5m_A,5w5m_B"
            ]
         },
        """
        retD = {}
        try:
            ind = -1
            if "chain_ids" in pLD:
                for ii, chS in enumerate(pLD["chain_ids"]):
                    if chainId in chS:
                        ind = ii
                        break
                if ind >= 0:
                    for ky in ["proteinName", "receptorType", "receptorDescription", "species"]:
                        if ky in pLD and len(pLD[ky]) > ind:
                            retD[ky] = pLD[ky][ind]
                else:
                    logger.info("missing chain %r in %r", chainId, pLD["chain_ids"])
            else:
                logger.info("missing chain details for %r in %r", chainId, pLD)
        except Exception as e:
            logger.exception("Failing for %r with %s", chainId, str(e))

        return retD

    def __getField(self, lineList, label):
        label = "IMGT chain description  "
        ret = None
        for line in lineList:
            if line.startswith(label):
                ret = line[len(label) :]
                break
        return ret

    def __splitDomains(self, pdbId, lineList):
        retD = {}
        startLabel1 = "-DOMAIN      IMGT domain description  "
        startLabel2 = "-LIKE-DOMAIN IMGT domain description  "
        startLabel3 = "-LIKE-DOMAIN IMGT domain description "
        #
        geneLabel1 = "-DOMAIN      IMGT gene and allele     "
        geneLabel2 = "-LIKE-DOMAIN IMGT gene and allele     "
        geneLabel3 = "-LIKE-DOMAIN IMGT gene and allele    "
        tD = {}
        domain = None
        numD = 0
        for line in lineList:
            if line[1:].startswith(startLabel1):
                numD += 1
                domain = line.split(" ")[0].strip()
                description = line[len(startLabel1) + 1 :].strip()
                continue
            if line[1:].startswith(startLabel2):
                numD += 1
                domain = line.split(" ")[0].strip()
                description = line[len(startLabel2) + 1 :].strip()
                continue
            if line[2:].startswith(startLabel3):
                numD += 1
                domain = line.split(" ")[0].strip()
                description = line[len(startLabel3) + 1 :].strip()
                continue
            if domain and line.startswith(domain):
                tD.setdefault((domain, description, numD), []).append(line)
        #
        qD = {}
        for (domain, description, numD), cL in tD.items():
            for line in cL:
                if line[1:].startswith(geneLabel1):
                    qD.setdefault((domain + "|" + description + "|" + str(numD)), []).append(line[len(geneLabel1) + 1 :])
                if line[1:].startswith(geneLabel2):
                    qD.setdefault((domain + "|" + description + "|" + str(numD)), []).append(line[len(geneLabel2) + 1 :])
                if line[2:].startswith(geneLabel3):
                    qD.setdefault((domain + "|" + description + "|" + str(numD)), []).append(line[len(geneLabel3) + 1 :])
        #
        # "Homo sapiens IGHG4*01 (96.4%), Homo sapiens IGHG4*03 (96.4%), Homo sapiens IGHG4*04 (96.4%)",
        for ky, cL in qD.items():
            logger.debug("cL %r", cL)
            tS = "".join(cL)
            tS = " ".join(tS.split())
            logger.debug("tS %r", tS)
            #  handle some missing commas in the raw data -
            tS = tS.replace(") ", "),")
            logger.debug("TAX> %r tS %r", pdbId, tS)
            gnSL = tS.split(",")
            gDL = []
            for gnS in gnSL:
                tL = gnS.strip().split()
                logger.debug("tL %r", tL)
                geneAllele = tL[-2]
                taxName = " ".join(tL[:-2])
                gDL.append({"taxName": taxName, "geneAllele": geneAllele})
            retD[ky] = {"geneAlleles": gDL}

        return retD

    def __getAlignment(self, pdbId, lineList):
        try:
            startPat = "Chain amino acid sequence"
            endPat1 = "-DOMAIN"
            endPat2 = "-LIKE-DOMAIN"
            aL = []
            keep = False
            for line in lineList:
                if line.startswith(startPat):
                    keep = True
                    continue
                if line[1:].startswith(endPat1):
                    break
                if line[1:].startswith(endPat2):
                    break
                if line[2:].startswith(endPat2):
                    break
                if keep:
                    aL.append(line)
            #
            sLine = "".join(aL[1::2])
            mLine = "".join(aL[0::2])
            # Lots of cases with (UNK) sequences where REM 410 format is corrupt --
            # if "(UNK)" in sLine:
            #    logger.error("%r unknown or modified residue in one-letter-code sequence %r", pdbId, sLine[:30] + "...")
            #    return {}
            ok, indD = self.__findMatchingGroups(mLine, startGroup="[", endGroup="]")
            if not ok:
                logger.error("%r determining alignment boundaries fails", pdbId)
                return {}
            pdbRangeL = []
            for iBeg, iEnd in indD.items():
                if iEnd - iBeg <= 3:
                    continue
                pdbRangeL.append({"begEntitySeqId": iBeg + 1, "endEntitySeqId": iEnd + 1})
            ok, indD = self.__findMatchingGroups(mLine, startGroup="(", endGroup=")")
            if not ok:
                logger.error("%r determining alignment boundaries fails", pdbId)
                return {}
            imgtRangeL = []
            try:
                for k, v in indD.items():
                    tS = mLine[k + 1 : v]
                    tL = tS.split("-")
                    iBeg = tL[0]
                    iEnd = tL[1]
                    imgtRangeL.append({"begIMGTSeqId": iBeg, "endIMGTSeqId": iEnd})
            except Exception as e:
                logger.error("%r parsing boundaries fails with %r for %r", pdbId, str(e), mLine[:30] + "...")
                return {}
            #
            alignMapDL = []
            for pdbD, imgtD in zip(pdbRangeL, imgtRangeL):
                dD = pdbD
                dD.update(imgtD)
                alignMapDL.append(dD)
            return {"mapping": mLine, "pdbSeq": sLine, "alignMapDL": alignMapDL}
        except Exception as e:
            logger.exception("Failing %r with %r with %s", pdbId, mLine, str(e))
        return {}

    def __findMatchingGroups(self, strIn, startGroup="[", endGroup="]"):
        retD = {}
        dStack = []
        ok = True
        try:
            for i, cS in enumerate(strIn):
                if cS == startGroup:
                    dStack.append(i)
                elif cS == endGroup:
                    if len(dStack) == 0:
                        logger.error("No matching closing group at position: %r", str(i))
                        ok = False
                    retD[dStack.pop()] = i

            if len(dStack) > 0:
                logger.error("No matching opening group at: %r", strIn(dStack.pop()))
                ok = False
        except Exception:
            pass
        return ok, retD



##
#  File:           PharosTargetProvider.py
#  Date:           11-Jun-2021 jdw
#
#  Updates:
#   27-Feb-2023 dwp Update mysql loading command
#   23-Mar-2023 aae Download sql file to separate dir
#    5-May-2023 aae Check cache files exist in testCache()
##
"""
Accessors for Pharos target assignments.
"""

import logging
import os.path
import time

from rcsb.utils.io.ExecUtils import ExecUtils
from rcsb.utils.io.FileUtil import FileUtil
from rcsb.utils.io.MarshalUtil import MarshalUtil
from rcsb.utils.io.StashableBase import StashableBase
from rcsb.utils.seq.UniProtIdMappingProvider import UniProtIdMappingProvider

logger = logging.getLogger(__name__)


class PharosTargetProvider(StashableBase):
    """Accessors for Pharos target assignments."""

    def __init__(self, **kwargs):
        #
        self.__cachePath = kwargs.get("cachePath", ".")
        self.__dirName = "Pharos-targets"
        super(PharosTargetProvider, self).__init__(self.__cachePath, [self.__dirName])
        self.__dirPath = os.path.join(self.__cachePath, self.__dirName)
        #
        # Folder for SQL dump
        self.__sqlDumpDirName = "Pharos-targets-sqldump"
        self.__sqlDumpDirPath = os.path.join(self.__cachePath, self.__sqlDumpDirName)
        #
        self.__mU = MarshalUtil(workPath=self.__dirPath)
        self.__pharosSelectedTables = ["drug_activity", "cmpd_activity", "target", "protein", "t2tc"]
        reloadDb = kwargs.get("reloadDb", False)
        fromDb = kwargs.get("fromDb", False)
        useCache = kwargs.get("useCache", False)
        pharosDumpUrl = kwargs.get("pharosDumpUrl", None)
        mysqlUser = kwargs.get("mysqlUser", None)
        mysqlPassword = kwargs.get("mysqlPassword", None)
        self.__version = None
        if reloadDb or fromDb:
            self.__reload(
                targetsPath=self.__dirPath,
                sqlPath=self.__sqlDumpDirPath,
                reloadDb=reloadDb,
                fromDb=fromDb,
                useCache=useCache,
                pharosDumpUrl=pharosDumpUrl,
                mysqlUser=mysqlUser,
                mysqlPassword=mysqlPassword
            )
        #

    def testCache(self):
        ok = True
        for tbl in self.__pharosSelectedTables:
            outPath = os.path.join(self.__dirPath, "%s.tdd" % tbl)
            ok = self.__mU.exists(outPath) and ok
            if not ok:
                break
        return ok

    def getVersion(self):
        return self.__version

    def __reload(self, targetsPath, sqlPath, reloadDb=False, fromDb=False, useCache=False, pharosDumpUrl=None, mysqlUser=None, mysqlPassword=None):
        startTime = time.time()
        pharosDumpUrl = pharosDumpUrl if pharosDumpUrl else "http://juniper.health.unm.edu/tcrd/download/latest.sql.gz"
        pharosReadmeUrl = "http://juniper.health.unm.edu/tcrd/download/latest.README"
        ok = False
        fU = FileUtil()
        pharosDumpFileName = fU.getFileName(pharosDumpUrl)
        pharosDumpPath = os.path.join(sqlPath, pharosDumpFileName)
        pharosUpdatePath = os.path.join(sqlPath, "pharos-update.sql")
        pharosReadmePath = os.path.join(targetsPath, "pharos-readme.txt")  # needed by PharosTargetActivityProvider
        logPath = os.path.join(sqlPath, "pharosLoad.log")
        #
        fU.mkdir(sqlPath)
        fU.mkdir(targetsPath)
        #

        exU = ExecUtils()
        #
        if reloadDb:
            logger.info("useCache %r pharosDumpPath %r", useCache, pharosDumpPath)
            if useCache and self.__mU.exists(pharosDumpPath):
                ok = True
            else:
                logger.info("Fetching url %s path %s", pharosDumpUrl, pharosDumpPath)
                ok1 = fU.get(pharosDumpUrl, pharosDumpPath)
                ok2 = fU.get(pharosReadmeUrl, pharosReadmePath)
                logger.info("Completed fetch (%r) at %s (%.4f seconds)", ok1 and ok2, time.strftime("%Y %m %d %H:%M:%S", time.localtime()), time.time() - startTime)
            # ---
            readmeLines = self.__mU.doImport(pharosReadmePath, fmt="list")
            self.__version = readmeLines[0].split(" ")[1][1:] if readmeLines else "6"
            # ---
            logger.info("Filtering SQL dump %r for selected tables %r", pharosDumpFileName, self.__pharosSelectedTables)
            doWrite = True
            # Note: the pharos dump file latest.sql.gz is not gzipped
            with open(pharosDumpPath, "r", encoding="utf-8") as ifh, open(pharosUpdatePath, "w", encoding="utf-8") as ofh:
                for line in ifh:
                    if line.startswith("-- Table structure for table"):
                        tN = line.split(" ")[-1][1:-2]
                        doWrite = True if tN in self.__pharosSelectedTables else False
                    if doWrite:
                        ofh.write(line)
            # ---
            ok = exU.run(
                "mysql",
                execArgList=["-v", "-u", mysqlUser, "--password=%s" % mysqlPassword, "-e", "create database if not exists tcrd6;"],
                outPath=logPath,
                outAppend=False,
                timeOut=None,
            )
            # ok = exU.run(
            #     "mysql",
            #     execArgList=["-u", mysqlUser, "--password=%s" % mysqlPassword, "tcrd6"],
            #     outPath=logPath,
            #     inpPath=pharosDumpPath,
            #     outAppend=True,
            #     timeOut=None,
            # )
            #
            # Load pharos data into MySQL and trap for hangup, interruption and termination signals (note that the SIG prefix isn't necessary).
            # (Also note that subprocess will use `/bin/sh`, which may point to `dash` and thus may behave differently than if run via `bash`)
            logger.info("Loading pharosUpdatePath %s into mysql db 'tcrd6' (check log: %s)", os.path.abspath(pharosUpdatePath), os.path.abspath(logPath))
            shellCmd = f'trap "" HUP INT TERM; nohup mysql -u {mysqlUser} --password={mysqlPassword} tcrd6 < {pharosUpdatePath}'
            ok = exU.runShell(
                shellCmd,
                outPath=logPath,
                inpPath=None,
                outAppend=True,
                timeOut=None,
            )
            logger.info("SQL dump restore status %r", ok)
        # --
        if fromDb:
            for tbl in self.__pharosSelectedTables:
                outPath = os.path.join(targetsPath, "%s.tdd" % tbl)
                # if useCache and self.__mU.exists(outPath):
                #   continue
                ok = exU.run(
                    "mysql",
                    execArgList=["-u", mysqlUser, "--password=%s" % mysqlPassword, "-e", "use tcrd6; select * from %s;" % tbl],
                    outPath=outPath,
                    outAppend=False,
                    timeOut=None,
                    suppressStderr=True,
                )
                logger.info("SQL table %s export status %r", tbl, ok)
        return ok

    def exportProteinFasta(self, fastaPath, taxonPath, addTaxonomy=False):
        try:
            proteinFilePath = os.path.join(self.__dirPath, "protein.tdd")
            pDL = self.__mU.doImport(proteinFilePath, fmt="tdd", rowFormat="dict")
            fD = {}
            taxonL = []
            if addTaxonomy:
                umP = UniProtIdMappingProvider(self.__cachePath)
                umP.reload(useCache=True)
                #
                for pD in pDL:
                    unpId = pD["uniprot"]
                    proteinId = pD["id"]
                    seq = pD["seq"]
                    taxId = umP.getMappedId(unpId, mapName="NCBI-taxon")
                    taxId = taxId if taxId else "-1"
                    cD = {"sequence": seq, "uniprotId": unpId, "proteinId": proteinId, "taxId": taxId}
                    seqId = ""
                    cL = []
                    for k, v in cD.items():
                        if k in ["sequence"]:
                            continue
                        cL.append(str(v))
                        cL.append(str(k))
                    seqId = "|".join(cL)
                    fD[seqId] = cD
                    taxonL.append("%s\t%s" % (seqId, taxId))
                ok = self.__mU.doExport(taxonPath, taxonL, fmt="list")
            else:
                for pD in pDL:
                    unpId = pD["uniprot"]
                    proteinId = pD["id"]
                    seq = pD["seq"]
                    cD = {"sequence": seq, "uniprotId": unpId, "proteinId": proteinId}
                    seqId = ""
                    cL = []
                    for k, v in cD.items():
                        if k in ["sequence"]:
                            continue
                        cL.append(str(v))
                        cL.append(str(k))
                    seqId = "|".join(cL)
                    fD[seqId] = cD
            #
            logger.info("Writing %d pharos targets to %s", len(fD), fastaPath)
            ok = self.__mU.doExport(fastaPath, fD, fmt="fasta", makeComment=True)
        except Exception as e:
            logger.exception("Failing with %s", str(e))
        return ok



##
#  File:           CARDTargetFeatureProvider.py
#  Date:           12-Jun-2021 jdw
#
#  Updates:
#   14-Mar-2023 dwp  CARD feature data will stop being used temporarily (in favor of CARD annotations)
##
"""
Accessors and generators for CARD target feature data.
"""

import datetime
import logging
import os.path
import time

from rcsb.utils.io.FileUtil import FileUtil
from rcsb.utils.io.MarshalUtil import MarshalUtil
from rcsb.utils.io.StashableBase import StashableBase
from rcsb.utils.targets.CARDTargetProvider import CARDTargetProvider

logger = logging.getLogger(__name__)


class CARDTargetFeatureProvider(StashableBase):
    """Accessors and generators for CARD target feature data."""

    def __init__(self, **kwargs):
        #
        self.__cachePath = kwargs.get("cachePath", ".")
        self.__dirName = "CARD-features"
        super(CARDTargetFeatureProvider, self).__init__(self.__cachePath, [self.__dirName])
        self.__dirPath = os.path.join(self.__cachePath, self.__dirName)
        #
        self.__mU = MarshalUtil(workPath=self.__dirPath)
        self.__fD = self.__reload(self.__dirPath, **kwargs)
        #

    def testCache(self, minCount=590):
        logger.info("CARD feature count %d", len(self.__fD["features"]) if "features" in self.__fD else 0)
        if self.__fD and "features" in self.__fD and len(self.__fD["features"]) > minCount:
            return True
        else:
            return False

    def hasFeatures(self, rcsbEntityId):
        return rcsbEntityId.upper() in self.__fD["features"]

    def getFeatures(self, rcsbEntityId):
        try:
            return self.__fD["features"][rcsbEntityId.upper()]
        except Exception:
            pass
        return []

    def reload(self):
        self.__fD = self.__reload(self.__dirPath, useCache=True)
        return True

    def __getFeatureDataPath(self):
        return os.path.join(self.__dirPath, "CARD-feature-data.json")

    def __reload(self, dirPath, **kwargs):
        startTime = time.time()
        fD = {}
        useCache = kwargs.get("useCache", True)
        ok = False
        featurePath = self.__getFeatureDataPath()
        #
        logger.info("useCache %r featurePath %r", useCache, featurePath)
        if useCache and self.__mU.exists(featurePath):
            fD = self.__mU.doImport(featurePath, fmt="json")
        else:
            fU = FileUtil()
            fU.mkdir(dirPath)
        # ---
        logger.info("Completed reload (%r) at %s (%.4f seconds)", ok, time.strftime("%Y %m %d %H:%M:%S", time.localtime()), time.time() - startTime)
        return fD

    def buildFeatureList(self, sequenceMatchFilePath, useTaxonomy=False):
        """Build polymer entity feature list for the matching entities in the input sequence match file.

        Args:
            sequenceMatchFilePath (str): sequence match output file path
            useTaxonomy (bool): apply taxonomy criteria to filter matches. Defaults to False

        Returns:
            bool: True for success or False otherwise


        """
        rDL = []
        cardP = CARDTargetProvider(cachePath=self.__cachePath, useCache=False)
        mD = self.__mU.doImport(sequenceMatchFilePath, fmt="json")
        #
        provenanceSource = "CARD"
        refScheme = "PDB entity"
        assignVersion = cardP.getAssignmentVersion()
        for queryId, matchDL in mD.items():
            qCmtD = self.__decodeComment(queryId)
            modelId = qCmtD["modelId"]
            if not cardP.hasModel(modelId):
                logger.info("Skipping CARD model %r", modelId)
                continue

            for matchD in matchDL:
                #
                fpL = []
                if "alignedRegions" in matchD:
                    fpL = [{"beg_seq_id": arD["targetBegin"], "end_seq_id": arD["targetEnd"]} for arD in matchD["alignedRegions"]]
                else:
                    fpL = [{"beg_seq_id": matchD["targetBegin"], "end_seq_id": matchD["targetEnd"]}]
                tCmtD = self.__decodeComment(matchD["target"])
                entryId = tCmtD["entityId"].split("_")[0]
                entityId = tCmtD["entityId"].split("_")[1]
                nm = cardP.getModelValue(modelId, "modelName")
                descr = cardP.getModelValue(modelId, "descr")
                featureId = cardP.getModelValue(modelId, "cvTermId")
                rD = {
                    "entry_id": entryId,
                    "entity_id": entityId,
                    "type": "CARD_MODEL",
                    "feature_id": featureId,
                    "name": nm,
                    "description": descr,
                    "provenance_source": provenanceSource,
                    "reference_scheme": refScheme,
                    "assignment_version": assignVersion,
                    "feature_positions": fpL,
                    "query_tax_name": matchD["queryTaxName"] if "queryTaxName" in matchD else None,
                    "target_tax_name": matchD["targetTaxName"] if "targetTaxName" in matchD else None,
                    "match_status": matchD["taxonomyMatchStatus"] if "taxonomyMatchStatus" in matchD else None,
                    "lca_tax_name": matchD["lcaTaxName"] if "lcaTaxName" in matchD else None,
                }
                rDL.append(rD)
        #
        qD = {}
        dD = {}
        for rD in rDL:
            eId = rD["entry_id"] + "_" + rD["entity_id"]
            fId = rD["feature_id"]
            if (eId, fId) in dD:
                continue
            dD[(eId, fId)] = True
            qD.setdefault(eId, []).append(rD)
        # --
        if useTaxonomy:
            fqD = {}
            for eId, rDL in qD.items():
                mL = []
                oL = []
                for rD in rDL:
                    if "match_status" not in rD:
                        continue
                    if rD["match_status"] in ["matched", "alternate strain", "alternate strain/serotype/isolate/genotype", "query is ancestor", "query is descendant"]:
                        mL.append(rD)
                    elif rD["match_status"] == "orthologous match (by lca)":
                        oL.append(rD)
                    else:
                        continue
                if not mL and not oL:
                    continue
                logger.debug("eId %r mL (%d) oL (%d)", eId, len(mL), len(oL))
                fqD.setdefault(eId, []).extend(mL if mL else oL)
        else:
            fqD = qD
        # --

        fp = self.__getFeatureDataPath()
        tS = datetime.datetime.now().isoformat()
        vS = datetime.datetime.now().strftime("%Y-%m-%d")
        ok = self.__mU.doExport(fp, {"version": vS, "created": tS, "features": fqD}, fmt="json", indent=3)
        return ok

    def __decodeComment(self, comment, separator="|"):
        dD = {}
        try:
            ti = iter(comment.split(separator))
            dD = {tup[1]: tup[0] for tup in zip(ti, ti)}
        except Exception:
            pass
        return dD



##
#  File:           CARDTargetProvider.py
#  Date:           27-Nov-2020 jdw
#
#  Updates:
#   14-Mar-2023 dwp  Add filterForHomologs() method to filter CARD data for only protein homolog models;
#                    Remove unused getTargetDataPath() and getCofactorDataPath() methods
##
"""
Accessors for CARD target assignments.

"""

import datetime
import logging
import os.path
import time

from rcsb.utils.io.FileUtil import FileUtil
from rcsb.utils.io.MarshalUtil import MarshalUtil

logger = logging.getLogger(__name__)


class CARDTargetProvider:
    """Accessors for CARD target assignments."""

    def __init__(self, **kwargs):
        #
        self.__cachePath = kwargs.get("cachePath", ".")
        self.__dirPath = os.path.join(self.__cachePath, "CARD-targets")
        #
        self.__mU = MarshalUtil(workPath=self.__dirPath)
        self.__oD, self.__version = self.__reload(self.__dirPath, **kwargs)
        #

    def testCache(self, minCount=3000):
        if self.__oD and len(self.__oD) > minCount:
            return True
        else:
            return False

    def hasModel(self, modelId):
        return modelId in self.__oD

    def getModelValue(self, modelId, key):
        try:
            return self.__oD[modelId][key]
        except Exception:
            return None

    def getAssignmentVersion(self):
        return self.__version

    def reload(self):
        self.__oD, self.__version = self.__reload(self.__dirPath, useCache=True)

    def __reload(self, dirPath, **kwargs):
        oD = None
        version = None
        startTime = time.time()
        useCache = kwargs.get("useCache", True)
        #
        ok = False
        fU = FileUtil()
        #
        # CARDDumpUrl = kwargs.get("CARDDumpUrl", "https://card.mcmaster.ca/latest/data/broadstreet-v3.1.0.tar.bz2")
        cardDumpUrl = kwargs.get("CARDDumpUrl", "https://card.mcmaster.ca/latest/data")
        cardDumpFileName = "card-data.tar.bz2"
        cardDumpPath = os.path.join(dirPath, cardDumpFileName)
        cardDumpDirPath = os.path.join(dirPath, "dump-targets")
        #
        fU.mkdir(dirPath)
        cardDataPath = os.path.join(dirPath, "card-select-data.json")
        #
        # ---
        # Load CARD data
        logger.info("useCache %r CARDDumpPath %r", useCache, cardDumpPath)
        if useCache and self.__mU.exists(cardDataPath):
            qD = self.__mU.doImport(cardDataPath, fmt="json")
            version = qD["version"]
            oD = qD["data"]
        else:
            logger.info("Fetching url %s path %s", cardDumpUrl, cardDumpPath)
            ok = fU.get(cardDumpUrl, cardDumpPath)
            fU.mkdir(cardDumpDirPath)
            fU.uncompress(cardDumpPath, outputDir=cardDumpDirPath)
            fU.unbundleTarfile(os.path.join(cardDumpDirPath, cardDumpFileName[:-4]), dirPath=cardDumpDirPath)
            logger.info("Completed fetch (%r) at %s (%.4f seconds)", ok, time.strftime("%Y %m %d %H:%M:%S", time.localtime()), time.time() - startTime)
            oD, version = self.__parseCardData(os.path.join(cardDumpDirPath, "card.json"))
            #
            # Filter for only protein homolog models (this can be removed if/when we decide to include all types of models)
            oD = self.__filterForHomologs(os.path.join(cardDumpDirPath, "protein_fasta_protein_homolog_model.fasta"), oD)
            #
            tS = datetime.datetime.now().isoformat()
            qD = {"version": version, "created": tS, "data": oD}
            oD = qD["data"]
            ok = self.__mU.doExport(cardDataPath, qD, fmt="json", indent=3)
            logger.info("Export CARD data (%d) status %r", len(oD), ok)
        # ---
        return oD, version

    def exportCardFasta(self, fastaPath, taxonPath):
        ok = self.__exportCardFasta(fastaPath, taxonPath, self.__oD)
        return ok

    def __exportCardFasta(self, fastaPath, taxonPath, cardD):
        """Export a CARD sequence target fasta file

        Args:
            fastaPath (str): fasta output file path
            cardD (dict): card selected data dictionary

        Returns:
            (bool): True for success or False otherwise
        """
        sD = {}
        taxonL = []
        try:
            for modelId, tD in cardD.items():
                modelBitScore = None
                # aroAcc = tD["accession"]
                aroId = tD["cvTermId"]
                if "sequences" not in tD:
                    continue
                modelBitScore = tD["modelBitScore"] if "modelBitScore" in tD else None
                for qD in tD["sequences"]:
                    sId = qD["seqId"]
                    seq = qD["sequence"]
                    taxId = qD["taxId"]
                    cD = {"sequence": seq, "modelId": modelId, "aroId": aroId, "seqId": sId, "taxId": taxId}
                    cD["bitScore"] = modelBitScore if modelBitScore else "-1.0"
                    #
                    cId = ""
                    cL = []
                    for k, v in cD.items():
                        if k in ["sequence"]:
                            continue
                        cL.append(str(v))
                        cL.append(str(k))
                    cId = "|".join(cL)
                    sD[cId] = cD
                    taxonL.append("%s\t%s" % (cId, taxId))

            ok = self.__mU.doExport(fastaPath, sD, fmt="fasta", makeComment=True)
            logger.info("Export CARD fasta (%d) status %r", len(sD), ok)
            ok = self.__mU.doExport(taxonPath, taxonL, fmt="list")
            logger.info("Export Taxon (%d) status %r", len(taxonL), ok)
        except Exception as e:
            logger.exception("Failing for model %r tD %r with %s", modelId, tD, str(e))
        return ok

    def __parseCardData(self, filePath):
        """Parse CARD target data

        Args:
            filePath (str): card json data file

        Returns:
            (dict, string): card selected data dictionary, card version string
        """
        try:
            oD = {}
            version = None
            cD = self.__mU.doImport(filePath, fmt="json")
            logger.info("CARD model count (%d)", len(cD))
            for modelId, mD in cD.items():
                if modelId.startswith("_"):
                    if modelId == "_version":
                        version = mD
                    continue
                oD[modelId] = {}  # modelId = '1028'
                for kTup in [
                    ("ARO_accession", "accession"),  # 'ARO_accession', '3001059'
                    ("ARO_id", "cvTermId"),  # 'ARO_id', '37439'
                    ("ARO_name", "name"),  # 'ARO_name', 'SHV-1'
                    ("ARO_description", "descr"),  # 'SHV-1 is a broad-spectrum beta-lactamase...
                    ("model_name", "modelName"),  # 'model_name', 'SHV-1'
                    ("model_type", "modelType"),  # 'model_type', 'protein homolog model'
                ]:
                    if kTup[0] in mD:
                        oD[modelId][kTup[1]] = mD[kTup[0]]
                # Add in category details (Family, Drug Class(es), and Resistance mechanism)
                acD = {}
                if "ARO_category" in mD:
                    aroCategoryD = mD["ARO_category"]  # 'ARO_category', large dict containing each type of 'category_aro_class_name' ("AMR Gene Family", "Drug Class", ...)
                    for cvId, catD in aroCategoryD.items():
                        if all([k in catD for k in ["category_aro_class_name", "category_aro_accession", "category_aro_name", "category_aro_description"]]):
                            if catD["category_aro_class_name"] == "AMR Gene Family":
                                acD["familyCvTermId"] = cvId
                                acD["familyAccession"] = catD["category_aro_accession"]
                                acD["familyName"] = catD["category_aro_name"]
                                acD["familyDescription"] = catD["category_aro_description"]
                            if catD["category_aro_class_name"] == "Drug Class":
                                acD.setdefault("drugClasses", []).append(catD["category_aro_name"])
                            if catD["category_aro_class_name"] == "Resistance Mechanism":
                                acD["resistanceMechanism"] = catD["category_aro_name"]
                oD[modelId].update(acD)
                #
                try:
                    if "model_sequences" in mD:
                        for seqId, tD in mD["model_sequences"]["sequence"].items():
                            oD[modelId].setdefault("sequences", []).append(
                                {"seqId": seqId, "sequence": tD["protein_sequence"]["sequence"], "taxId": tD["NCBI_taxonomy"]["NCBI_taxonomy_id"]}
                            )
                except Exception as e:
                    logger.exception("Failing with %s", str(e))

                try:
                    if "model_param" in mD and "blastp_bit_score" in mD["model_param"] and "param_value" in mD["model_param"]["blastp_bit_score"]:
                        oD[modelId]["modelBitScore"] = mD["model_param"]["blastp_bit_score"]["param_value"]

                except Exception as e:
                    logger.exception("Failing with %s", str(e))

        except Exception as e:
            logger.exception("Failing with %s", str(e))
        return oD, version

    def __filterForHomologs(self, filePath, targetD):
        """Filter the CARD target data to select only for protein homologs

        Args:
            filePath (str): path to CARD fasta protein homolog model file (protein_fasta_protein_homolog_model.fasta; from source)
            targetD (dict): dictionary generated from __parseCardData (i.e., oD)

        Returns:
            (dict): filtered card selected data dictionary
        """

        with open(filePath, "r", encoding="utf-8") as f:
            data = f.readlines()

        aroL = [i.split("|")[2].strip("ARO:") for i in data if i.startswith(">")]

        filteredD = {}
        for k, v in targetD.items():
            if v["accession"] in aroL:
                filteredD.update({k: v})

        return filteredD



##
#  File:           SAbDabTargetFeatureProvider.py
#  Date:           11-Jun-2021 jdw
#
#  Updated:
#
##
"""
Accessors for Thera-SAbDab(Therapeutic Structural Antibody Database) target features.
"""

import datetime
import logging
import os.path
import time

from rcsb.utils.io.FileUtil import FileUtil
from rcsb.utils.io.MarshalUtil import MarshalUtil
from rcsb.utils.io.StashableBase import StashableBase
from rcsb.utils.targets.SAbDabTargetProvider import SAbDabTargetProvider

logger = logging.getLogger(__name__)


class SAbDabTargetFeatureProvider(StashableBase):
    """Accessors for Thera-SAbDab(Therapeutic Structural Antibody Database) target features."""

    # Link out using the INN therapeutic name -
    # http://opig.stats.ox.ac.uk/webapps/newsabdab/therasabdab/search/?therapeutic=Coltuximab
    def __init__(self, **kwargs):
        #
        self.__cachePath = kwargs.get("cachePath", ".")
        useCache = kwargs.get("useCache", True)
        self.__dirName = "SAbDab-features"
        super(SAbDabTargetFeatureProvider, self).__init__(self.__cachePath, [self.__dirName])
        self.__dirPath = os.path.join(self.__cachePath, self.__dirName)
        #
        self.__mU = MarshalUtil(workPath=self.__dirPath)
        self.__fD = self.__reload(self.__dirPath, useCache)
        #

    def testCache(self, minCount=500):
        logger.info(
            "Therapeutic SAbDab feature count %d Assignment count %d",
            len(self.__fD["features"]) if "features" in self.__fD else 0,
            len(self.__fD["assignments"]) if "assignments" in self.__fD else 0,
        )
        if self.__fD and "features" in self.__fD and len(self.__fD["features"]) > minCount and self.__fD and "assignments" in self.__fD and len(self.__fD["assignments"]):
            return True
        else:
            return False

    def getVersion(self):
        try:
            return self.__fD["version"]
        except Exception:
            pass
        return None

    def hasFeatures(self, rcsbEntityId):
        return rcsbEntityId.upper() in self.__fD["features"]

    def getFeatures(self, rcsbEntityId):
        try:
            return self.__fD["features"][rcsbEntityId.upper()]
        except Exception:
            return []

    def getAssignment(self, instanceId, featureKey):
        """Return the value of the key feature for the input instance identifier.

        Args:
            instanceId (str): instance identifier '<pdbId>.<authAsymId>'
            featureKey (str): assignment feature key: one of pdb|Hchain|Lchain|model|antigen_chain|antigen_type|
                              antigen_het_name|antigen_name|heavy_subclass|light_subclass|light_ctype)

        Returns:
            str:  feature value or None
        """
        fVal = None
        try:
            fVal = self.__fD["assignments"][instanceId][featureKey]
        except Exception:
            fVal = None
        return fVal

    def hasAssignment(self, instanceId):
        """Return if assignment data is available for the input instance.

        Args:
            instanceId (str): instance identifier '<pdbId>.<authAsymId>'

        Returns:
            bool: True for success or False otherwise
        """
        return instanceId in self.__fD["assignments"]

    def __getFeatureDataPath(self):
        return os.path.join(self.__dirPath, "sabdab-feature-data.json")

    def reload(self):
        self.__fD = self.__reload(self.__dirPath, useCache=True)
        return True

    def __reload(self, dirPath, useCache):
        startTime = time.time()
        fD = {}

        ok = False
        featurePath = self.__getFeatureDataPath()
        #
        logger.info("useCache %r featurePath %r", useCache, featurePath)
        if useCache and self.__mU.exists(featurePath):
            fD = self.__mU.doImport(featurePath, fmt="json")
        else:
            fU = FileUtil()
            fU.mkdir(dirPath)
        # ---
        logger.info("Completed reload (%r) at %s (%.4f seconds)", ok, time.strftime("%Y %m %d %H:%M:%S", time.localtime()), time.time() - startTime)
        return fD

    def buildFeatureList(self, sequenceMatchFilePath):
        """Build polymer entity feature list for the matching entities in the input sequence match file.

        Args:
            sequenceMatchFilePath (str): sequence match output file path

        Returns:
            bool: True for success or False otherwise

        """
        rDL = []
        stP = SAbDabTargetProvider(cachePath=self.__cachePath, useCache=False)
        if not stP.testCache():
            logger.warning("Skipping build of polymer entity feature list because SAbDab Target data is missing.")
            return False
        mD = self.__mU.doImport(sequenceMatchFilePath, fmt="json")
        #
        provenanceSource = "SAbDab"
        refScheme = "PDB entity"
        assignVersion = stP.getAssignmentVersion()
        #
        # - sort out if we match light and heavy chains
        #
        iD = {}
        fullMatchD = {}
        for queryId, matchDL in mD.items():
            qCmtD = self.__decodeComment(queryId)
            # Tanezumab|therapeutic|light|chain
            thName = qCmtD["therapeutic"]
            chainType = qCmtD["chain"]
            for matchD in matchDL:
                tCmtD = self.__decodeComment(matchD["target"])
                entryId = tCmtD["entityId"].split("_")[0]
                entityId = tCmtD["entityId"].split("_")[1]
                iD[(thName, chainType, entryId)] = entityId
        logger.info("Match index length (%d)", len(iD))
        for (thName, chainType, entryId), entityId in iD.items():
            if chainType == "light":
                continue
            if (thName, "light", entryId) in iD:
                fullMatchD[(thName, "heavy", entryId, entityId)] = True
                lEntityId = iD[(thName, "light", entryId)]
                fullMatchD[(thName, "light", entryId, lEntityId)] = True
        logger.info("Antibody entity match length (%d)", len(fullMatchD))
        #
        # - Add features for full matches -
        for queryId, matchDL in mD.items():
            qCmtD = self.__decodeComment(queryId)
            # Tanezumab|therapeutic|light|chain
            thName = qCmtD["therapeutic"]
            chainType = qCmtD["chain"]
            #
            for matchD in matchDL:
                fpL = []
                if "alignedRegions" in matchD:
                    fpL = [{"beg_seq_id": arD["targetBegin"], "end_seq_id": arD["targetEnd"]} for arD in matchD["alignedRegions"]]
                else:
                    fpL = [{"beg_seq_id": matchD["targetBegin"], "end_seq_id": matchD["targetEnd"]}]
                #
                tCmtD = self.__decodeComment(matchD["target"])
                entryId = tCmtD["entityId"].split("_")[0]
                entityId = tCmtD["entityId"].split("_")[1]
                if (thName, chainType, entryId, entityId) not in fullMatchD:
                    continue
                ii = 1
                for fType, fKy in [
                    ("SABDAB_ANTIBODY_NAME", "antibodyName"),
                    ("SABDAB_ANTIBODY_FORMAT", "antibodyFormat"),
                    ("SABDAB_ANTIBODY_CH1_ISOTYPE", "ch1IsoType"),
                    ("SABDAB_ANTIBODY_LIGHT_CHAIN_TYPE", "VD_LC"),
                    ("SABDAB_ANTIBODY_TARGET", "target"),
                ]:
                    if fType == "Antibody_Light_Chain_Type" and chainType == "heavy":
                        continue
                    fVL = stP.getFeatures(thName, fKy)
                    if not fVL:
                        continue
                    for fV in fVL:
                        rD = {
                            "entry_id": entryId,
                            "entity_id": entityId,
                            "type": fType,
                            "feature_id": thName + "_" + chainType + "_" + str(ii),
                            "name": fV,
                            "provenance_source": provenanceSource,
                            "reference_scheme": refScheme,
                            "assignment_version": assignVersion,
                            "feature_positions": fpL,
                        }
                        rDL.append(rD)
                        ii += 1
        #
        qD = {}
        for rD in rDL:
            eId = rD["entry_id"] + "_" + rD["entity_id"]
            qD.setdefault(eId, []).append(rD)
        #
        logger.info("Antibody matches (%d)", len(qD))
        #
        fp = self.__getFeatureDataPath()
        tS = datetime.datetime.now().isoformat()
        vS = datetime.datetime.now().strftime("%Y-%m-%d")
        ok = self.__mU.doExport(fp, {"version": vS, "created": tS, "features": qD, "assignments": stP.getAssignments()}, fmt="json", indent=3)
        return ok

    def __decodeComment(self, comment, separator="|"):
        dD = {}
        try:
            ti = iter(comment.split(separator))
            dD = {tup[1]: tup[0] for tup in zip(ti, ti)}
        except Exception:
            pass
        return dD



##
#  File:           ChEMBLTargetActivityProvider.py
#  Date:           9-Nov-2020 jdw
#
#  Updated:
#   9-Feb-2023 aae  Update ChEMBL baseVersion to 31
#  18-Jul-2023 dwp  Update ChEMBL baseVersion to 33
##
"""
Accessors for ChEMBL target activity data.

"""

import datetime
import logging
import os.path
import time

from chembl_webresource_client.settings import Settings

# pylint: disable=ungrouped-imports
try:
    from chembl_webresource_client.new_client import new_client  # fails when service is down
except Exception:
    pass

from rcsb.utils.multiproc.MultiProcUtil import MultiProcUtil
from rcsb.utils.io.FileUtil import FileUtil
from rcsb.utils.io.MarshalUtil import MarshalUtil
from rcsb.utils.io.StashableBase import StashableBase
from rcsb.utils.io.UrlRequestUtil import UrlRequestUtil


Settings.Instance().TIMEOUT = 10  # pylint: disable=no-member
Settings.Instance().MAX_LIMIT = 50  # pylint: disable=no-member
Settings.MAX_LIMIT = 50

logger = logging.getLogger(__name__)


class ChEMBLTargetActivityWorker(object):
    """A skeleton worker class that implements the interface expected by the multiprocessing module
    for fetching ChEMBL activity data --
    """

    def __init__(self, **kwargs):
        _ = kwargs

    def fetchActivity(self, dataList, procName, optionsD, workingDir):
        """Fetch ChEMBL activity for the input ChEMBL target identifier list"""
        _ = workingDir
        successList = []
        failList = []
        retList = []
        diagList = []
        #
        try:
            chunkSize = optionsD.get("chunkSize", 50)
            atL = optionsD.get("attributeList", [])
            maxActivity = optionsD.get("maxActivity", None)

            for ii in range(0, len(dataList), chunkSize):
                logger.info("Begin chunk at ii %d/(for %d targets)", ii, len(dataList))
                try:
                    act = new_client.activity  # pylint: disable=no-member
                    act.set_format("json")
                    actDL = (
                        act.filter(target_chembl_id__in=dataList[ii : ii + chunkSize])
                        .filter(standard_type__in=["IC50", "Ki", "EC50", "Kd"])
                        .filter(standard_value__isnull=False)
                        .order_by("-standard_value")
                        .only(atL)
                    )
                    logger.info("Results (%d)", len(actDL))
                    if actDL:
                        for actD in actDL[:maxActivity] if maxActivity else actDL:
                            actD = self.__activitySelect(atL, actD)
                            actD["molecule_name"], actD["inchi_key"], _ = self.__getMoleculeDetails(actD["molecule_chembl_id"])
                            actD["action"], actD["moa"], actD["max_phase"] = self.__getMechanismDetails(actD["molecule_chembl_id"])
                            retList.append((actD["target_chembl_id"], actD))
                except Exception as e:
                    logger.exception("Failing for chunk starting at %d with %s", ii, str(e)[:200])
            successList = sorted(set(dataList) - set(failList))
            if failList:
                logger.info("%s returns %d definitions with failures: %r", procName, len(failList), failList)

            logger.debug("%s built target interactions for %d/%d entries failures %d", procName, len(retList), len(dataList), len(failList))
        except Exception as e:
            logger.exception("Failing %s for %d data items %s", procName, len(dataList), str(e))
        #
        return successList, retList, diagList

    def __activitySelect(self, atL, aD):
        return {at: aD[at] if at in aD else None for at in atL}

    def __getMoleculeDetails(self, chemblId):
        name = inchiKey = smiles = None
        try:
            atL = ["pref_name", "molecule_structures"]
            molecule = new_client.molecule  # pylint: disable=no-member
            molecule.set_format("json")
            tD = molecule.filter(molecule_chembl_id__exact=chemblId).only(atL)[0]
            name = tD["pref_name"] if tD and "pref_name" in tD else None
            if tD and "molecule_structures" in tD and tD["molecule_structures"]:
                smiles = tD["molecule_structures"]["canonical_smiles"] if tD and "molecule_structures" in tD and "canonical_smiles" in tD["molecule_structures"] else None
                inchiKey = tD["molecule_structures"]["standard_inchi_key"] if tD and "molecule_structures" in tD and "standard_inchi_key" in tD["molecule_structures"] else None
        except Exception as e:
            logger.exception("Failing for %s with %s", chemblId, str(e))
        return name, inchiKey, smiles

    def __getMechanismDetails(self, chemblId):
        actionType = moa = maxPhase = None
        try:
            atL = ["action_type", "mechanism_of_action", "max_phase"]
            mechanism = new_client.mechanism  # pylint: disable=no-member
            mechanism.set_format("json")
            tD = mechanism.filter(molecule_chembl_id__exact=chemblId).only(atL)[0]
            actionType = tD["action_type"] if tD and "action_type" in tD else None
            moa = tD["mechanism_of_action"] if tD and "mechanism_of_action" in tD else None
            maxPhase = tD["max_phase"] if tD and "max_phase" in tD else None
        except Exception as e:
            logger.exception("Failing for %s with %s", chemblId, str(e))
        return actionType, moa, maxPhase


class ChEMBLTargetActivityProvider(StashableBase):
    """Accessors for ChEMBL target activity data."""

    def __init__(self, cachePath, useCache):
        #
        self.__cachePath = cachePath
        self.__dirName = "ChEMBL-target-activity"
        super(ChEMBLTargetActivityProvider, self).__init__(self.__cachePath, [self.__dirName])
        self.__dirPath = os.path.join(self.__cachePath, self.__dirName)
        self.__mU = MarshalUtil(workPath=self.__cachePath)
        baseVersion = 33
        self.__version = baseVersion
        logger.info("ChEMBL API MAX_LIMIT %r", Settings.Instance().MAX_LIMIT)  # pylint: disable=no-member
        self.__aD, self.__allIdD = self.__reload(self.__dirPath, useCache)

    def testCache(self, minCount=1):
        if self.__aD and (len(self.__aD) >= minCount):
            logger.info("Activity data cached for (%d) targets", len(self.__aD))
            return True
        return False

    def getAssignmentVersion(self):
        return self.__version

    def getTargetActivityDataPath(self):
        return os.path.join(self.__dirPath, "chembl-target-activity-data.json")

    def reload(self):
        self.__aD, self.__allIdD = self.__reload(self.__dirPath, True)

    def __reload(self, dirPath, useCache):
        startTime = time.time()
        aD = {}
        allIdD = {}
        fU = FileUtil()
        fU.mkdir(dirPath)
        targetActivityFilePath = self.getTargetActivityDataPath()
        #
        if useCache and fU.exists(targetActivityFilePath):
            logger.info("useCache %r using %r", useCache, targetActivityFilePath)
            qD = self.__mU.doImport(targetActivityFilePath, fmt="json")
            aD = qD["activity"] if "activity" in qD else {}
            idL = qD["all_ids"] if "all_ids" in qD else []
            allIdD = {k: k in aD for k in idL}
        #
        logger.info(
            "Completed reload (%d activities) (%d tried identifiers) at %s (%.4f seconds)",
            len(aD),
            len(allIdD),
            time.strftime("%Y %m %d %H:%M:%S", time.localtime()),
            time.time() - startTime,
        )
        #
        return aD, allIdD

    def getTargetActivity(self, targetChEMBLId):
        try:
            return self.__aD[targetChEMBLId] if targetChEMBLId in self.__aD else []
        except Exception:
            return []

    def hasTargetActivity(self, targetChEMBLId):
        try:
            return targetChEMBLId in self.__aD
        except Exception:
            return False

    def getTargetIdList(self, sequenceMatchFilePath):
        chemblIdList = []
        try:
            mD = self.__mU.doImport(sequenceMatchFilePath, fmt="json")
            # --- cofactor list
            chemblIdList = []
            for queryId in mD:
                qCmtD = self.__decodeComment(queryId)
                tS = qCmtD["chemblId"]
                tL = tS.split(",")
                chemblIdList.extend(tL)
            chemblIdList = list(set(chemblIdList))
            logger.info("Total targets from sequence matching (%d)", len(chemblIdList))
        except Exception as e:
            logger.exception("Failing for %r with %s", sequenceMatchFilePath, str(e))
        return chemblIdList

    def fetchTargetActivityData(self, targetChEMBLIdList, skip="none", maxActivity=10, chunkSize=50):
        """Get cofactor activity data for the input ChEMBL target list.

        Args:
            targetChEMBLIdList (list): list of ChEMBL target identifiers
            skip (str, optional): Skip searching identifiers previously tried|matched|none (default: none)
            chunkSize (int, optional): ChEMBL API batch size for fetches (default: 50)
            maxActivity (int, optional): maximum number of activity records to return per target (default: 10)

        Returns:
          bool:  True for success or False otherwise

        """
        atL = [
            "assay_chembl_id",
            "assay_description",
            "assay_type",
            "canonical_smiles",
            "ligand_efficiency",
            "molecule_chembl_id",
            "parent_molecule_chembl_id",
            "pchembl_value",
            "standard_relation",
            "standard_type",
            "standard_units",
            "standard_value",
            "target_chembl_id",
        ]
        ok = False
        targetD = self.__aD if self.__aD else {}
        logger.info("Fetching activities for starting target list (%d) skip option %r", len(targetChEMBLIdList), skip)
        idList = []
        if skip == "matched":
            for tId in targetChEMBLIdList:
                if tId in self.__aD:
                    continue
                idList.append(tId)
        elif skip == "tried":
            for tId in targetChEMBLIdList:
                if tId in self.__allIdD:
                    continue
                idList.append(tId)
        else:
            idList = targetChEMBLIdList

        numToProcess = len(idList)
        logger.info("Filtered target list (%d)", len(idList))
        try:
            for ii in range(0, len(idList), chunkSize):
                logger.info("Begin chunk at ii %d/%d", ii, numToProcess)
                try:
                    act = new_client.activity  # pylint: disable=no-member
                    act.set_format("json")
                    actDL = (
                        act.filter(target_chembl_id__in=idList[ii : ii + chunkSize])
                        .filter(standard_type__in=["IC50", "Ki", "EC50", "Kd"])
                        .filter(standard_value__isnull=False)
                        .order_by("-standard_value")
                        .only(atL)
                    )
                    logger.info("Results for index %d (%d)", ii, len(actDL))
                    if actDL:
                        for actD in actDL[:maxActivity] if maxActivity else actDL:
                            actD = self.__activitySelect(atL, actD)
                            actD["molecule_name"], actD["inchi_key"], _ = self.getMoleculeDetails(actD["molecule_chembl_id"])
                            actD["action_type"], actD["moa"], actD["max_phase"] = self.getMechanismDetails(actD["molecule_chembl_id"])
                            targetD.setdefault(actD["target_chembl_id"], []).append(actD)
                    #
                    logger.info("Completed chunk starting at (%d)", ii)
                    tS = datetime.datetime.now().isoformat()
                    vS = datetime.datetime.now().strftime("%Y-%m-%d")
                    ok = self.__mU.doExport(self.getTargetActivityDataPath(), {"version": vS, "created": tS, "activity": targetD, "all_ids": idList}, fmt="json", indent=3)
                    logger.info("Wrote completed chunk starting at (%d) (%r)", ii, ok)
                #
                except Exception as e:
                    logger.exception("Failing with chunk at index %d with %s", ii, str(e)[:200])
        #
        except Exception as e:
            logger.exception("Failing with %s", str(e))
        return ok

    def __activitySelect(self, atL, aD):
        return {at: aD[at] if at in aD else None for at in atL}

    def getMoleculeDetails(self, chemblId):
        name = inchiKey = smiles = None
        try:
            atL = ["pref_name", "molecule_structures"]
            molecule = new_client.molecule  # pylint: disable=no-member
            molecule.set_format("json")
            tD = molecule.filter(molecule_chembl_id__exact=chemblId).only(atL)[0]
            name = tD["pref_name"] if tD and "pref_name" in tD else None
            if tD and "molecule_structures" in tD and tD["molecule_structures"]:
                smiles = tD["molecule_structures"]["canonical_smiles"] if tD and "molecule_structures" in tD and "canonical_smiles" in tD["molecule_structures"] else None
                inchiKey = tD["molecule_structures"]["standard_inchi_key"] if tD and "molecule_structures" in tD and "standard_inchi_key" in tD["molecule_structures"] else None
        except Exception as e:
            logger.exception("Failing for %s with %s", chemblId, str(e))
        return name, inchiKey, smiles

    def getStatusDetails(self):
        try:
            version = releaseDateString = None
            baseUrl = "https://www.ebi.ac.uk"
            endPoint = "chembl/api/data/status.json"
            hL = []
            pD = {}
            ureq = UrlRequestUtil()
            ret, retCode = ureq.get(baseUrl, endPoint, pD, headers=hL, returnContentType="JSON")
            logger.info("retCode %r ret %r", retCode, ret)
            if ret:
                tS = ret["chembl_db_version"] if ret and "chembl_db_version" in ret else None
                version = tS.split("_")[1] if tS and tS.split("_")[1] else None
                releaseDateString = ret["chembl_release_date"] if "chembl_release_date" in ret else None
        except Exception as e:
            logger.exception("Failing with %s", str(e))
        return version, releaseDateString

    def getMechanismDetails(self, chemblId):
        actionType = moa = maxPhase = None
        try:
            atL = ["action_type", "mechanism_of_action", "max_phase"]
            mechanism = new_client.mechanism  # pylint: disable=no-member
            mechanism.set_format("json")
            tD = mechanism.filter(molecule_chembl_id__exact=chemblId).only(atL)[0]
            actionType = tD["action_type"] if tD and "action_type" in tD else None
            moa = tD["mechanism_of_action"] if tD and "mechanism_of_action" in tD else None
            maxPhase = tD["max_phase"] if tD and "max_phase" in tD else None
        except Exception as e:
            logger.exception("Failing for %s with %s", chemblId, str(e))
        return actionType, moa, maxPhase

    def fetchTargetActivityDataMulti(self, targetChEMBLIdList, skip="none", maxActivity=10, chunkSize=50, numProc=4):
        """Get cofactor activity data for the input ChEMBL target list (multiprocessing mode).

        Args:
            targetChEMBLIdList (list): list of ChEMBL target identifiers
            skip (str, optional): Skip searching identifiers previously tried|matched|none (default: none)
            chunkSize(int, optional): ChEMBL API outer batch size for fetches (default: 50)
            numProc (int, optional): number processes to invoke. (default: 4)
            maxActivity (int, optional): number of activity records to return per target. (default: 10)

        Returns:
          bool:  True for success or False otherwise

        """
        atL = [
            "assay_chembl_id",
            "assay_description",
            "assay_type",
            "canonical_smiles",
            "ligand_efficiency",
            "molecule_chembl_id",
            "parent_molecule_chembl_id",
            "pchembl_value",
            "standard_relation",
            "standard_type",
            "standard_units",
            "standard_value",
            "target_chembl_id",
        ]
        logger.info("Fetching activities for starting target list (%d) skip option %r", len(targetChEMBLIdList), skip)
        ok = False
        targetD = self.__aD if self.__aD else {}
        idList = []
        if skip == "matched":
            for tId in targetChEMBLIdList:
                if tId in self.__aD:
                    continue
                idList.append(tId)
        elif skip == "tried":
            for tId in targetChEMBLIdList:
                if tId in self.__allIdD:
                    continue
                idList.append(tId)
        else:
            idList = targetChEMBLIdList

        numToProcess = len(idList)
        allIdL = []
        logger.info("Filtered target list (%d)", numToProcess)
        ok = numToProcess == 0
        try:
            for ii in range(0, len(idList), chunkSize):
                logger.info("Begin outer chunk at ii %d (total targets %d)", ii, numToProcess)
                tIdList = idList[ii : ii + chunkSize]
                #
                tD = self.__getActivityMulti(tIdList, atL, maxActivity=maxActivity, numProc=numProc, chunkSize=5)
                targetD.update(tD)
                #
                allIdL.extend(tIdList)
                #
                logger.info("Completed outer chunk (%d) total processed targets (%d/%d)", ii, len(allIdL), numToProcess)
                tS = datetime.datetime.now().isoformat()
                vS = datetime.datetime.now().strftime("%Y-%m-%d")
                ok = self.__mU.doExport(self.getTargetActivityDataPath(), {"version": vS, "created": tS, "activity": targetD, "all_ids": allIdL}, fmt="json", indent=3)
                logger.info("Wrote completed outer chunk starting at (%d) (%r)", ii, ok)
        except Exception as e:
            logger.exception("Failing with %s", str(e))
        return ok

    def __getActivityMulti(self, idList, atL, maxActivity=None, numProc=2, chunkSize=5):
        """ """
        rD = {}
        ctaW = ChEMBLTargetActivityWorker()
        mpu = MultiProcUtil(verbose=True)
        optD = {"attributeList": atL, "chunkSize": chunkSize, "maxActivity": maxActivity}
        mpu.setOptions(optD)
        mpu.set(workerObj=ctaW, workerMethod="fetchActivity")
        ok, failList, resultList, _ = mpu.runMulti(dataList=idList, numProc=numProc, numResults=1, chunkSize=chunkSize)
        if failList:
            logger.info("Target Id activity failures (%d): %r", len(failList), failList)
        #
        for (targetId, actD) in resultList[0]:
            rD.setdefault(targetId, []).append(actD)
        #
        logger.info("Completed with multi-proc status %r failures %r total targets with data (%d)", ok, len(failList), len(rD))
        return rD

    def __decodeComment(self, comment, separator="|"):
        dD = {}
        try:
            ti = iter(comment.split(separator))
            dD = {tup[1]: tup[0] for tup in zip(ti, ti)}
        except Exception:
            pass
        return dD
